#include "HDC1080.h"
#include "HDC1080_PlatformSpecific.h"

#include <stdint.h>
#include <stddef.h>

static HDC1080_Configuration currentConfig;
static int isMeasurementTriggered;
static int isMeasurementTemperature;

HDC1080_Status HDC1080_Init() {
	HDC1080_Status status;

	status = HDC1080_PlatformSpecific_Init();
	if (status) {
		return status;
	}

	status = HDC1080_GetConfiguration(&currentConfig);
	if (status) {
		return status;
	}

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_Deinit() {
	return HDC1080_PlatformSpecific_Deinit();
}

static HDC1080_Status HDC1080_ReadRegister(uint8_t registerAddress, uint16_t* registerValue) {
	HDC1080_Status status;

	uint8_t buffer[2];

	if (isMeasurementTriggered) {
		return HDC1080_Status_InvalidOperation;
	}

	status = HDC1080_PlatformSpecific_BeginRead(registerAddress);
	if (status) {
		return status;
	}

	status = HDC1080_PlatformSpecific_EndRead(buffer, sizeof(buffer));
	if (status) {
		return status;
	}

	*registerValue =
	(((uint16_t)buffer[0]) << 8) |
	(((uint16_t)buffer[1]) << 0);

	return HDC1080_Status_Ok;
}

static HDC1080_Status HDC1080_EndReadRegisters(uint16_t* registers, size_t registersSize) {
	HDC1080_Status status;

	uint8_t* buffer = (uint8_t*)registers;

	status = HDC1080_PlatformSpecific_EndRead(buffer, registersSize);
	if (status) {
		return status;
	}

	// efficiently swap bytes on little endian platforms
	for (size_t i = 0; i < registersSize / sizeof(uint16_t); i++) {
		uint16_t registerValue =
		(((uint16_t)buffer[i * 2 + 0]) << 8) |
		(((uint16_t)buffer[i * 2 + 1]) << 0);

		registers[i] = registerValue;
	}

	return HDC1080_Status_Ok;
}

static HDC1080_Status HDC1080_WriteRegister(uint8_t registerAddress, uint16_t registerValue) {
	HDC1080_Status status;

	uint8_t buffer[2];

	buffer[0] = (registerValue & 0xFF00) >> 8;
	buffer[1] = registerValue & 0xFF;

	status = HDC1080_PlatformSpecific_Write(registerAddress, buffer, sizeof(buffer));
	if (status) {
		return status;
	}

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_GetDefaultConfiguration(HDC1080_Configuration* config) {
	config->enableHeater = 0;
	config->mode = HDC1080_Mode_MeasureTemperatureThenHumidity;
	config->temperatureResolution = HDC1080_TemperatureResolution_14bit;
	config->humidityResolution = HDC1080_HumidityResolution_14bit;

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_GetConfiguration(HDC1080_Configuration* config) {
	HDC1080_Status status;
	uint16_t configRegVal;

	status = HDC1080_ReadRegister(HDC1080_REG_CONFIG, &configRegVal);
	if (status) {
		return status;
	}

	config->enableHeater = HDC1080_GET_FIELD(HDC1080_CONFIG_HEAT_FIELD, configRegVal);
	config->mode = HDC1080_GET_FIELD(HDC1080_CONFIG_MODE_FIELD, configRegVal);
	config->temperatureResolution = HDC1080_GET_FIELD(HDC1080_CONFIG_TRES_FIELD, configRegVal);
	config->humidityResolution = HDC1080_GET_FIELD(HDC1080_CONFIG_HRES_FIELD, configRegVal);

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_SetConfiguration(HDC1080_Configuration* config) {
	HDC1080_Status status;

	uint16_t configRegVal =
	HDC1080_SET_FIELD(HDC1080_CONFIG_HEAT_FIELD, config->enableHeater) |
	HDC1080_SET_FIELD(HDC1080_CONFIG_MODE_FIELD, config->mode) |
	HDC1080_SET_FIELD(HDC1080_CONFIG_TRES_FIELD, config->temperatureResolution) |
	HDC1080_SET_FIELD(HDC1080_CONFIG_HRES_FIELD, config->humidityResolution);

	status = HDC1080_WriteRegister(HDC1080_REG_CONFIG, configRegVal);
	if (status) {
		return status;
	}

	currentConfig.enableHeater = config->enableHeater;
	currentConfig.mode = config->mode;
	currentConfig.temperatureResolution = config->temperatureResolution;
	currentConfig.humidityResolution = config->humidityResolution;

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_TriggerTemperatureAndHumidityConversion() {
	HDC1080_Status status;

	if (currentConfig.mode != HDC1080_Mode_MeasureTemperatureThenHumidity) {
		return HDC1080_Status_InvalidOperation;
	}

	if (isMeasurementTriggered) {
		return HDC1080_Status_InvalidOperation;
	}

	status = HDC1080_PlatformSpecific_BeginRead(HDC1080_REG_TEMP);
	if (status) {
		return status;
	}

	isMeasurementTriggered = 1;

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_TriggerTemperatureConversion() {
	HDC1080_Status status;

	if (currentConfig.mode != HDC1080_Mode_MeasureTemperatureOrHumidity) {
		return HDC1080_Status_InvalidOperation;
	}

	if (isMeasurementTriggered) {
		return HDC1080_Status_InvalidOperation;
	}

	status = HDC1080_PlatformSpecific_BeginRead(HDC1080_REG_TEMP);
	if (status) {
		return status;
	}

	isMeasurementTriggered = 1;
	isMeasurementTemperature = 1;

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_TriggerHumidityConversion() {
	HDC1080_Status status;

	if (currentConfig.mode != HDC1080_Mode_MeasureTemperatureOrHumidity) {
		return HDC1080_Status_InvalidOperation;
	}

	if (isMeasurementTriggered) {
		return HDC1080_Status_InvalidOperation;
	}

	status = HDC1080_PlatformSpecific_BeginRead(HDC1080_REG_HUMI);
	if (status) {
		return status;
	}

	isMeasurementTriggered = 1;
	isMeasurementTemperature = 0;

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_GetRawTemperature(uint16_t* temperature) {
	if (currentConfig.mode != HDC1080_Mode_MeasureTemperatureOrHumidity) {
		return HDC1080_Status_InvalidOperation;
	}

	if (!isMeasurementTriggered || !isMeasurementTemperature) {
		return HDC1080_Status_InvalidOperation;
	}

	isMeasurementTriggered = 0;

	return HDC1080_EndReadRegisters(temperature, sizeof(*temperature));
}

HDC1080_Status HDC1080_GetRawHumidity(uint16_t* humidity) {
	if (currentConfig.mode != HDC1080_Mode_MeasureTemperatureOrHumidity) {
		return HDC1080_Status_InvalidOperation;
	}

	if (!isMeasurementTriggered || isMeasurementTemperature) {
		return HDC1080_Status_InvalidOperation;
	}

	isMeasurementTriggered = 0;

	return HDC1080_EndReadRegisters(humidity, sizeof(*humidity));
}

HDC1080_Status HDC1080_GetRawTemperatureAndHumidity(uint16_t* temperature, uint16_t* humidity) {
	HDC1080_Status status;
	uint16_t regs[2];

	if (currentConfig.mode != HDC1080_Mode_MeasureTemperatureThenHumidity) {
		return HDC1080_Status_InvalidOperation;
	}

	if (!isMeasurementTriggered) {
		return HDC1080_Status_InvalidOperation;
	}

	isMeasurementTriggered = 0;

	status = HDC1080_EndReadRegisters(regs, sizeof(regs));
	if (status) {
		return status;
	}

	*temperature = regs[0];
	*humidity = regs[1];

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_GetTemperature(float* temperature) {
	HDC1080_Status status;
	uint16_t raw;

	status = HDC1080_GetRawTemperature(&raw);
	if (status) {
		return status;
	}

	*temperature = ((int)raw) * 165 / (float)(1 << 16) - 40;

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_GetHumidity(float* humidity) {
	HDC1080_Status status;
	uint16_t raw;

	status = HDC1080_GetRawHumidity(&raw);
	if (status) {
		return status;
	}

	*humidity = ((int)raw) * 100 / (float)(1 << 16);

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_GetTemperatureAndHumidity(float* temperature, float* humidity) {
	HDC1080_Status status;
	uint16_t rawTemperature;
	uint16_t rawHumidity;

	status = HDC1080_GetRawTemperatureAndHumidity(&rawTemperature, &rawHumidity);
	if (status) {
		return status;
	}

	*temperature = ((int)rawTemperature) * 165 / (float)(1 << 16) - 40;
	*humidity = ((int)rawHumidity) * 100 / (float)(1 << 16);

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_EnableHeater() {
	if (currentConfig.enableHeater) {
		return HDC1080_Status_Ok;
	}

	HDC1080_Configuration newConfig;
	newConfig.enableHeater = 1;
	newConfig.mode = currentConfig.mode;
	newConfig.humidityResolution = currentConfig.humidityResolution;
	newConfig.temperatureResolution = currentConfig.temperatureResolution;

	return HDC1080_SetConfiguration(&newConfig);
}

HDC1080_Status HDC1080_DisableHeater() {
	if (!currentConfig.enableHeater) {
		return HDC1080_Status_Ok;
	}

	HDC1080_Configuration newConfig;
	newConfig.enableHeater = 1;
	newConfig.mode = currentConfig.mode;
	newConfig.humidityResolution = currentConfig.humidityResolution;
	newConfig.temperatureResolution = currentConfig.temperatureResolution;

	return HDC1080_SetConfiguration(&newConfig);
}

HDC1080_Status HDC1080_GetBatteryStatus(HDC1080_BatteryStatus* batterySatus) {
	HDC1080_Status status;
	uint16_t configRegVal;

	status = HDC1080_ReadRegister(HDC1080_REG_CONFIG, &configRegVal);
	if (status) {
		return status;
	}

	*batterySatus = HDC1080_GET_FIELD(HDC1080_CONFIG_BTST_FIELD, configRegVal);

	return HDC1080_Status_Ok;
}

HDC1080_Status HDC1080_GetManufacturerNumber(uint16_t* manufacturerNumber) {
	return HDC1080_ReadRegister(HDC1080_REG_IDMFG, manufacturerNumber);
}

HDC1080_Status HDC1080_GetDeviceId(uint16_t* deviceId) {
	return HDC1080_ReadRegister(HDC1080_REG_IDPROD, deviceId);
}

HDC1080_Status HDC1080_GetSerialNumber(uint64_t* serialNumber) {
	HDC1080_Status status;

	uint16_t reg[3];
	status = HDC1080_ReadRegister(HDC1080_REG_ID1, reg + 0);
	if (status) {
		return status;
	}

	status = HDC1080_ReadRegister(HDC1080_REG_ID2, reg + 1);
	if (status) {
		return status;
	}

	status = HDC1080_ReadRegister(HDC1080_REG_ID3, reg + 2);
	if (status) {
		return status;
	}

	*serialNumber =
	((uint64_t)(reg[0]) << 25) |
	((uint64_t)(reg[1]) << 9) |
	((uint64_t)(reg[2]) >> 7);

	return HDC1080_Status_Ok;

}